/* SIMPLE - Simple Is a Macro Processing Language Element */
/* Copyright (c) 1998 David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * This file contains the definitions of the builtin functions.
 */

#include <stdio.h>
#include <stdlib.h>

#include "simple.h"
#include "token.h"
#include "stack.h"

int
readinteger(token_list *tl)
{
  char *tmp;
  int i;
  tmp=tokenlisttostring(tl);
  if (sscanf(tmp,"%d",&i)<1)
    fatal(0,"This is not an integer: %s",tmp);
  FREE(tmp);
  return i;
}

void
writeinteger(int i,token_list *tl)
{
  char *tmp;
  int j;
  tmp=malloc(30);
#if defined(HAVE_SNPRINTF)
  if (snprintf(tmp,30,"%d",i)==30)
    fatal(1,"Ooops! not enough characters...");
#else
  sprintf(tmp,"%d",i);
#endif
  for (j=0;tmp[j];j++)
    appendtoken(tl,tmp[j]);
  FREE(tmp);
}

void
bi_identity(token_list *expansion,int argc,token_list *argv)
{
  if (argc>=2)
    copylist(expansion,argv+1);
}

void
bi_void(token_list *expansion,int argc,token_list *argv)
{
}

void
bi_output(token_list *expansion,int argc,token_list *argv)
{
  int i;
  if (argc<2) return;
  for (i=0;i<argv[1].size;i++)
    outputtoken(argv[1].l[i]);
}

void
bi_define(token_list *expansion,int argc,token_list *argv)
{
  char *name;
  command todef;
  if (argc!=3)
    fatal(0,"<def> expects two arguments.");
  name=tokenlisttostring(argv+1);
  todef=identifycommand(name);
  FREE(name);
  if (todef<0)
    fatal(0,"Trying to define something you shouldn't.");
  defineuser(argv+2,todef);
}

void
bi_include(token_list *expansion,int argc,token_list *argv)
{
  char *name;
  if (argc!=2)
    fatal(0,"<include> expects one argument.");
  name=tokenlisttostring(argv+1);
  pushinputfile(name);
  FREE(name);
}

void
bi_it(token_list *expansion,int argc,token_list *argv)
{
  callmacro(expansion,argc-1,argv+1);
  /* Perhaps we should do this nonrecursively - but then, one is
   * simply _not_ supposed to call <|||||||||||||id|Hello!>... */
}

void
bi_quote(token_list *expansion,int argc,token_list *argv)
{
  int i;
  if (argc!=2)
    fatal(0,"<quote> expects one argument.");
  ensurelength(expansion,argv[1].size*2);
  expansion->size=argv[1].size*2;
  for (i=0;i<argv[1].size;i++) {
    expansion->l[2*i]=QUOTE_NEXT;
    expansion->l[2*i+1]=argv[1].l[i];
  }
}

void
bi_dquote(token_list *expansion,int argc,token_list *argv)
{
  int i;
  if (argc!=2)
    fatal(0,"<dquote> expects one argument.");
  ensurelength(expansion,argv[1].size*4);
  expansion->size=argv[1].size*4;
  for (i=0;i<argv[1].size;i++) {
    expansion->l[4*i]=QUOTE_NEXT;
    expansion->l[4*i+1]=QUOTE_NEXT;
    expansion->l[4*i+2]=QUOTE_NEXT;
    expansion->l[4*i+3]=argv[1].l[i];
  }
}

void
bi_inputform(token_list *expansion,int argc,token_list *argv)
{
  token tk;
  int i;
  if (argc!=2)
    fatal(0,"<inputform> expects one argument.");
  for (i=0;i<argv[1].size;i++) {
    tk=argv[1].l[i];
    if (tk<FIRST_SPECIAL) {
      if ((tk==char_escape_next)||(tk==char_escape_string)||
	  (tk==char_comment)||(tk==char_begin_command)||
	  (tk==char_end_command)||(tk==char_next_param)||
	  (tk==char_open_quote)||(tk==char_close_quote)||
	  (tk==char_quote_next)) {
	if (char_escape_next==UNREACHABLE) fatal(0,"Unreachable character.");
	appendtoken(expansion,char_escape_next);
      }
      appendtoken(expansion,tk);
    } else if (tk==BEGIN_COMMAND) {
      if (char_begin_command==UNREACHABLE) fatal(0,"Unreachable character.");
      appendtoken(expansion,char_begin_command);
    } else if (tk==END_COMMAND) {
      if (char_end_command==UNREACHABLE) fatal(0,"Unreachable character.");
      appendtoken(expansion,char_end_command);
    } else if (tk==NEXT_PARAM) {
      if (char_next_param==UNREACHABLE) fatal(0,"Unreachable character.");
      appendtoken(expansion,char_next_param);
    } else if (tk==OPEN_QUOTE) {
      if (char_open_quote==UNREACHABLE) fatal(0,"Unreachable character.");
      appendtoken(expansion,char_open_quote);
    } else if (tk==CLOSE_QUOTE) {
      if (char_close_quote==UNREACHABLE) fatal(0,"Unreachable character.");
      appendtoken(expansion,char_close_quote);
    } else if (tk==QUOTE_NEXT) {
      if (char_quote_next==UNREACHABLE) fatal(0,"Unreachable character.");
      appendtoken(expansion,char_quote_next);
    }
  }
}

void
bi_if(token_list *expansion,int argc,token_list *argv)
{
  int i;
  if (argc%3==0)
    fatal(0,"<if> can't deal with %d parameters.",argc-1);
  for (i=0;3*i+3<argc;i++) {
    if (comparelist(argv+(3*i+1),argv+(3*i+2))) {
      copylist(expansion,argv+(3*i+3));
      return;
    }
  }
  if (3*i+1<argc) {
    copylist(expansion,argv+(3*i+1));
  } /* Otherwise expansion is empty */
}

void
bi_quit(token_list *expansion,int argc,token_list *argv)
{
  outputtoken(EOF); /* The proper way to quit. */
}

void
bi_error(token_list *expansion,int argc,token_list *argv)
{
  char *errmsg;
  if (argc>=2) {
    errmsg=tokenlisttostring(argv+1);
    fatal(0,"%s",errmsg);
    FREE(errmsg); /* This doesn't get executed, of course. */
  } else
    fatal(0,"User error.");
}

void
bi_head(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=2)
    fatal(0,"<head> expects one argument.");
  if (argv[1].size==0)
    fatal(0,"<head> of an empty list is invalid.");
  appendtoken(expansion,argv[1].l[0]);
}

void
bi_tail(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=2)
    fatal(0,"<tail> expects one argument.");
  if (argv[1].size==0)
    fatal(0,"<tail> of an empty list is invalid.");
  copylist(expansion,argv+1);
  removefirst(expansion);
}

void
bi_ahead(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=2)
    fatal(0,"<ahead> expects one argument.");
  if (argv[1].size==0)
    fatal(0,"<ahead> of an empty list is invalid.");
  appendtoken(expansion,argv[1].l[argv[1].size-1]);
}

void
bi_atail(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=2)
    fatal(0,"<atail> expects one argument.");
  if (argv[1].size==0)
    fatal(0,"<atail> of an empty list is invalid.");
  copylist(expansion,argv+1);
  removelast(expansion);
}

void
bi_len(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=2)
    fatal(0,"<len> expects one argument.");
  writeinteger(argv[1].size,expansion);
}

void
bi_push(token_list *expansion,int argc,token_list *argv)
{
  char *name;
  token_stack *ts;
  int i;
  if (argc<=2)
    fatal(0,"<push> expects at least one argument.");
  name=tokenlisttostring(argv+1);
  ts=identifystack(name);
  FREE(name);
  for (i=2;i<argc;i++)
    pushonstack(ts,argv+i);
}

void
bi_last(token_list *expansion,int argc,token_list *argv)
{
  char *name;
  token_stack *ts;
  if (argc!=2)
    fatal(0,"<last> expects one argument.");
  name=tokenlisttostring(argv+1);
  ts=identifystack(name);
  FREE(name);
  lastfromstack(ts,expansion);
}

void
bi_pop(token_list *expansion,int argc,token_list *argv)
{
  char *name;
  token_stack *ts;
  if (argc!=2)
    fatal(0,"<pop> expects one argument.");
  name=tokenlisttostring(argv+1);
  ts=identifystack(name);
  FREE(name);
  popfromstack(ts);
}

void
bi_poplast(token_list *expansion,int argc,token_list *argv)
{
  char *name;
  token_stack *ts;
  if (argc!=2)
    fatal(0,"<poplast> expects one argument.");
  name=tokenlisttostring(argv+1);
  ts=identifystack(name);
  FREE(name);
  lastfromstack(ts,expansion);
  popfromstack(ts);
}

void
bi_depth(token_list *expansion,int argc,token_list *argv)
{
  char *name;
  token_stack *ts;
  if (argc!=2)
    fatal(0,"<depth> expects one argument.");
  name=tokenlisttostring(argv+1);
  ts=identifystack(name);
  FREE(name);
  writeinteger(stackdepth(ts),expansion);
}

void
bi_plus(token_list *expansion,int argc,token_list *argv)
{
  int i,j;
  j=0;
  for (i=1;i<argc;i++)
    j+=readinteger(argv+i);
  writeinteger(j,expansion);
}

void
bi_minus(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=3)
    fatal(0,"<-> expects two arguments.");
  writeinteger(readinteger(argv+1)-readinteger(argv+2),expansion);
}

void
bi_times(token_list *expansion,int argc,token_list *argv)
{
  int i,j;
  j=1;
  for (i=1;i<argc;i++)
    j*=readinteger(argv+i);
  writeinteger(j,expansion);
}

void
bi_div(token_list *expansion,int argc,token_list *argv)
{
  int d;
  if (argc!=3)
    fatal(0,"<div> expects two arguments.");
  d=readinteger(argv+2);
  if (d==0)
    fatal(0,"Division by zero.");
  writeinteger(readinteger(argv+1)/d,expansion);
}

void
bi_mod(token_list *expansion,int argc,token_list *argv)
{
  int d;
  if (argc!=3)
    fatal(0,"<mod> expects two arguments.");
  d=readinteger(argv+2);
  if (d==0)
    fatal(0,"Division by zero.");
  writeinteger(readinteger(argv+1)%d,expansion);
}

void
bi_eq(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=3)
    fatal(0,"<eq> expects two arguments.");
  writeinteger(readinteger(argv+1)==readinteger(argv+2),expansion);
}

void
bi_neq(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=3)
    fatal(0,"<neq> expects two arguments.");
  writeinteger(readinteger(argv+1)!=readinteger(argv+2),expansion);
}

void
bi_gt(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=3)
    fatal(0,"<gt> expects two arguments.");
  writeinteger(readinteger(argv+1)>readinteger(argv+2),expansion);
}

void
bi_ge(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=3)
    fatal(0,"<ge> expects two arguments.");
  writeinteger(readinteger(argv+1)>=readinteger(argv+2),expansion);
}

void
bi_lt(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=3)
    fatal(0,"<lt> expects two arguments.");
  writeinteger(readinteger(argv+1)<readinteger(argv+2),expansion);
}

void
bi_le(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=3)
    fatal(0,"<le> expects two arguments.");
  writeinteger(readinteger(argv+1)<=readinteger(argv+2),expansion);
}

void
bi_and(token_list *expansion,int argc,token_list *argv)
{
  int i,j;
  j=1;
  for (i=1;i<argc;i++)
    j=j&&readinteger(argv+i);
  writeinteger(j,expansion);
}

void
bi_or(token_list *expansion,int argc,token_list *argv)
{
  int i,j;
  j=0;
  for (i=1;i<argc;i++)
    j=j||readinteger(argv+i);
  writeinteger(j,expansion);
}

void
bi_not(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=2)
    fatal(0,"<not> expects one arguments.");
  writeinteger(!readinteger(argv+1),expansion);
}

void
bi_band(token_list *expansion,int argc,token_list *argv)
{
  int i,j;
  j=-1;
  for (i=1;i<argc;i++)
    j=j&readinteger(argv+i);
  writeinteger(j,expansion);
}

void
bi_bor(token_list *expansion,int argc,token_list *argv)
{
  int i,j;
  j=0;
  for (i=1;i<argc;i++)
    j=j|readinteger(argv+i);
  writeinteger(j,expansion);
}

void
bi_bnot(token_list *expansion,int argc,token_list *argv)
{
  if (argc!=2)
    fatal(0,"<bnot> expects one arguments.");
  writeinteger(~readinteger(argv+1),expansion);
}

void
bi_change_escape_next_char(token_list *expansion,int argc,token_list *argv)
{
  char *s;
  if (argc!=2)
    fatal(0,"<change_escape_next_char> expects one argument.");
  s=tokenlisttostring(argv+1);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_escape_next=s[0];
  else char_escape_next=UNREACHABLE;
  FREE(s);
}

void
bi_change_escape_string_char(token_list *expansion,int argc,token_list *argv)
{
  char *s;
  if (argc!=2)
    fatal(0,"<change_escape_string_char> expects one argument.");
  s=tokenlisttostring(argv+1);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_escape_string=s[0];
  else char_escape_string=UNREACHABLE;
  FREE(s);
}

void
bi_change_comment_char(token_list *expansion,int argc,token_list *argv)
{
  char *s;
  if (argc!=2)
    fatal(0,"<change_comment_char> expects one argument.");
  s=tokenlisttostring(argv+1);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_comment=s[0];
  else char_comment=UNREACHABLE;
  FREE(s);
}

void
bi_change_command_chars(token_list *expansion,int argc,token_list *argv)
{
  char *s;
  if (argc!=4)
    fatal(0,"<change_command_chars> expects three argument.");
  s=tokenlisttostring(argv+1);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_begin_command=s[0];
  else char_begin_command=UNREACHABLE;
  FREE(s);
  s=tokenlisttostring(argv+2);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_end_command=s[0];
  else char_end_command=UNREACHABLE;
  FREE(s);
  s=tokenlisttostring(argv+3);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_next_param=s[0];
  else char_next_param=UNREACHABLE;
  FREE(s);
}

void
bi_change_quote_chars(token_list *expansion,int argc,token_list *argv)
{
  char *s;
  if (argc!=4)
    fatal(0,"<change_quote_chars> expects three argument.");
  s=tokenlisttostring(argv+1);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_open_quote=s[0];
  else char_open_quote=UNREACHABLE;
  FREE(s);
  s=tokenlisttostring(argv+2);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_close_quote=s[0];
  else char_close_quote=UNREACHABLE;
  FREE(s);
  s=tokenlisttostring(argv+3);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_quote_next=s[0];
  else char_quote_next=UNREACHABLE;
  FREE(s);
}

void
bi_change_arg_char(token_list *expansion,int argc,token_list *argv)
{
  char *s;
  if (argc!=2)
    fatal(0,"<change_arg_char> expects one argument.");
  s=tokenlisttostring(argv+1);
  if (strlen(s)>1)
    fatal(0,"Single character argument expected.");
  if (s[0]!=0) char_at_sign=s[0];
  else char_at_sign=UNREACHABLE;
  FREE(s);
}

static struct {
  command who;
  void (* what)(token_list *,int,token_list *);
} builtin_table[] = {
  {BI_IDENTITY, bi_identity},
  {BI_VOID, bi_void},
  {BI_OUTPUT, bi_output},
  {BI_DEFINE, bi_define},
  {BI_INCLUDE, bi_include},
  {BI_IT, bi_it},
  {BI_QUOTE, bi_quote},
  {BI_DQUOTE, bi_dquote},
  {BI_INPUTFORM, bi_inputform},
  {BI_IF, bi_if},
  {BI_QUIT, bi_quit},
  {BI_ERROR, bi_error},
  {BI_HEAD, bi_head},
  {BI_TAIL, bi_tail},
  {BI_AHEAD, bi_ahead},
  {BI_ATAIL, bi_atail},
  {BI_LEN, bi_len},
  {BI_PUSH, bi_push},
  {BI_LAST, bi_last},
  {BI_POP, bi_pop},
  {BI_POPLAST, bi_poplast},
  {BI_DEPTH, bi_depth},
  {BI_PLUS, bi_plus},
  {BI_MINUS, bi_minus},
  {BI_TIMES, bi_times},
  {BI_DIV, bi_div},
  {BI_MOD, bi_mod},
  {BI_EQ, bi_eq},
  {BI_NEQ, bi_neq},
  {BI_GT, bi_gt},
  {BI_GE, bi_ge},
  {BI_LT, bi_lt},
  {BI_LE, bi_le},
  {BI_AND, bi_and},
  {BI_OR, bi_or},
  {BI_NOT, bi_not},
  {BI_BAND, bi_band},
  {BI_BOR, bi_bor},
  {BI_BNOT, bi_bnot},
  {BI_CHANGE_ESCAPE_NEXT_CHAR, bi_change_escape_next_char},
  {BI_CHANGE_ESCAPE_STRING_CHAR, bi_change_escape_string_char},
  {BI_CHANGE_COMMENT_CHAR, bi_change_comment_char},
  {BI_CHANGE_COMMAND_CHARS, bi_change_command_chars},
  {BI_CHANGE_QUOTE_CHARS, bi_change_quote_chars},
  {BI_CHANGE_ARG_CHAR, bi_change_arg_char},
  {0, NULL}
};

void
callbuiltin(token_list *expansion,command cmd,int argc,token_list *argv)
     /* Expand builtin token tk looking up its definition, and put the
      * result in expansion, the number of arguments being argc and
      * their values given by argv. */
{
  int i;
  for (i=0;builtin_table[i].who;i++) {
    if (cmd==builtin_table[i].who) {
      builtin_table[i].what(expansion,argc,argv);
      return;
    }
  }
  fatal(1,"Unresolved builtin!");
}
